@profile
def main():
    from torch_snippets import logger
    from torch_snippets import loader
    from torch_snippets import paths
    from torch_snippets import markup
    from torch_snippets import inspector
    from torch_snippets import load_defaults
    from torch_snippets import pdf_loader
    from torch_snippets import markup2
    from torch_snippets import registry
    from torch_snippets import ipython
    from torch_snippets import decorators
    from torch_snippets import misc
    from torch_snippets import dates

    # from torch_snippets import torch_loader
    from torch_snippets import s3_loader
    from torch_snippets import zen


print(main())
