# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {
    "np_2_b64": "adapters.ipynb",
    "b64_2_np": "adapters.ipynb",
    "b64_2_pdf": "adapters.ipynb",
    "parse_sz": "bokeh_plotting.ipynb",
    "get_bplot": "bokeh_plotting.ipynb",
    "BB": "bounding_boxes.ipynb",
    "df2bbs": "bounding_boxes.ipynb",
    "bbs2df": "bounding_boxes.ipynb",
    "bbfy": "bounding_boxes.ipynb",
    "jitter": "bounding_boxes.ipynb",
    "compute_eps": "bounding_boxes.ipynb",
    "enlarge_bbs": "bounding_boxes.ipynb",
    "shrink_bbs": "bounding_boxes.ipynb",
    "randint": "bounding_boxes.ipynb",
    "iou": "bounding_boxes.ipynb",
    "split_bb_to_xyXY": "bounding_boxes.ipynb",
    "combine_xyXY_to_bb": "bounding_boxes.ipynb",
    "to_relative": "bounding_boxes.ipynb",
    "to_absolute": "bounding_boxes.ipynb",
    "to": "capsule.ipynb",
    "train": "capsule.ipynb",
    "validate": "capsule.ipynb",
    "predict": "capsule.ipynb",
    "Capsule": "capsule.ipynb",
    "alt": "charts.ipynb",
    "Chart": "charts.ipynb",
    "confusion_matrix": "charts.ipynb",
    "CM": "charts.ipynb",
    "do": "imgaug_loader.ipynb",
    "bw": "imgaug_loader.ipynb",
    "rotate": "imgaug_loader.ipynb",
    "pad": "imgaug_loader.ipynb",
    "get_size": "imgaug_loader.ipynb",
    "rescale": "imgaug_loader.ipynb",
    "crop": "imgaug_loader.ipynb",
    "imgaugbbs2bbs": "imgaug_loader.ipynb",
    "bbs2imgaugbbs": "imgaug_loader.ipynb",
    "inspect": "inspector.ipynb",
    "is_in_notebook": "jupyter_notebook.ipynb",
    "ifexists": "load_defautls.ipynb",
    "ifnone": "load_defautls.ipynb",
    "get_console": "logging.ipynb",
    "console": "logging.ipynb",
    "print": "logging.ipynb",
    "render": "logging.ipynb",
    "logger": "logging.ipynb",
    "Debug": "logging.ipynb",
    "Info": "logging.ipynb",
    "Warn": "logging.ipynb",
    "Excep": "logging.ipynb",
    "reset_logger_width": "logging.ipynb",
    "isnamedtupleinstance": "markups.ipynb",
    "unpack": "markups.ipynb",
    "read_json": "markups.ipynb",
    "write_json": "markups.ipynb",
    "AttrDict": "registry.ipynb",
    "pretty_json": "markups.ipynb",
    "write_jsonl": "markups.ipynb",
    "read_jsonl": "markups.ipynb",
    "read_yaml": "markups.ipynb",
    "write_yaml": "markups.ipynb",
    "read_xml": "markups.ipynb",
    "write_xml": "markups.ipynb",
    "Config": "registry.ipynb",
    "input_to_str": "paths.ipynb",
    "output_to_path": "paths.ipynb",
    "size": "paths.ipynb",
    "extn": "paths.ipynb",
    "Glob": "paths.ipynb",
    "sample": "paths.ipynb",
    "mv": "paths.ipynb",
    "cp": "paths.ipynb",
    "rm": "paths.ipynb",
    "P": "paths.ipynb",
    "P.ls": "paths.ipynb",
    "P.__repr__": "paths.ipynb",
    "isdir": "paths.ipynb",
    "makedir": "paths.ipynb",
    "fname": "paths.ipynb",
    "fname2": "paths.ipynb",
    "stem": "paths.ipynb",
    "stems": "paths.ipynb",
    "parent": "paths.ipynb",
    "find": "paths.ipynb",
    "P.stems": "paths.ipynb",
    "zip_files": "paths.ipynb",
    "unzip_file": "paths.ipynb",
    "md5": "paths.ipynb",
    "remove_duplicates": "paths.ipynb",
    "readlines": "paths.ipynb",
    "read_lines": "paths.ipynb",
    "writelines": "paths.ipynb",
    "write_lines": "paths.ipynb",
    "rename_batch": "paths.ipynb",
    "dumpdill": "paths.ipynb",
    "loaddill": "paths.ipynb",
    "dill": "paths.ipynb",
    "PDF": "pdf.ipynb",
    "parse": "registry.ipynb",
    "parse_and_resolve": "registry.ipynb",
    "parse_string": "registry.ipynb",
    "registry": "registry.ipynb",
    "train_test_split": "sklegos.ipynb",
    "ColumnSelector": "sklegos.ipynb",
    "transform": "sklegos.ipynb",
    "GroupedEstimator": "sklegos.ipynb",
    "GroupedPredictor": "sklegos.ipynb",
    "EstimatorTransformer": "sklegos.ipynb",
    "MakeFrame": "sklegos.ipynb",
    "ImputeMissingValues": "sklegos.ipynb",
    "LambdaTransformer": "sklegos.ipynb",
    "Cat2Num": "sklegos.ipynb",
    "SplitDateColumn": "sklegos.ipynb",
}

modules = [
    "adapters.py",
    "bokeh_loader.py",
    "bb_utils.py",
    "trainer/capsule.py",
    "charts.py",
    "imgaug_loader.py",
    "inspector.py",
    "ipython.py",
    "load_defaults.py",
    "logger.py",
    "markup.py",
    "paths.py",
    "pdf_loader.py",
    "registry.py",
    "sklegos.py",
]

doc_url = "https://sizhky.github.io/torch_snippets/"

git_url = "https://github.com/sizhky/torch_snippets/tree/master/"


def custom_doc_links(name):
    return None
